@@ -1,21 +1,20 @@
-from __future__ import absolute_import
-from __future__ import print_function
+"""Command line argument parsing for cwltool."""
+from __future__ import absolute_import, print_function
 
 import argparse
-import logging
 import os
+from typing import (Any, AnyStr, Dict, List, MutableMapping, MutableSequence,
+                    Optional, Sequence, Union, cast)
 
-from typing import (Any, AnyStr, Dict, List, Sequence, Text, Union, cast)
-
-from . import loghandler
 from schema_salad.ref_resolver import file_uri
-from .process import (Process, shortname)
+from typing_extensions import Text  # pylint: disable=unused-import
+# move to a regular typing import when Python 3.3-3.6 is no longer supported
+
+from .loghandler import _logger
+from .process import Process, shortname  # pylint: disable=unused-import
 from .resolver import ga4gh_tool_registries
-from .software_requirements import (SOFTWARE_REQUIREMENTS_ENABLED)
-
-_logger = logging.getLogger("cwltool")
-
-DEFAULT_TMP_PREFIX = "tmp"
+from .software_requirements import SOFTWARE_REQUIREMENTS_ENABLED
+from .utils import DEFAULT_TMP_PREFIX
 
 
 def arg_parser():  # type: () -> argparse.ArgumentParser
@@ -26,49 +25,46 @@ def arg_parser():  # type: () -> argparse.ArgumentParser
                         help="Output directory, default current directory")
 
     parser.add_argument("--parallel", action="store_true", default=False,
-                        help="[experimental] Run jobs in parallel. "
-                             "Does not currently keep track of ResourceRequirements like the number of cores"
-                             "or memory and can overload this system")
+                        help="[experimental] Run jobs in parallel. ")
     envgroup = parser.add_mutually_exclusive_group()
     envgroup.add_argument("--preserve-environment", type=Text, action="append",
-                        help="Preserve specific environment variable when "
-                        "running CommandLineTools.  May be provided multiple "
-                        "times.", metavar="ENVVAR", default=["PATH"],
-                        dest="preserve_environment")
+                          help="Preserve specific environment variable when "
+                          "running CommandLineTools.  May be provided multiple "
+                          "times.", metavar="ENVVAR", default=["PATH"],
+                          dest="preserve_environment")
     envgroup.add_argument("--preserve-entire-environment", action="store_true",
-                        help="Preserve all environment variable when running "
-                        "CommandLineTools.", default=False,
-                        dest="preserve_entire_environment")
+                          help="Preserve all environment variable when running "
+                          "CommandLineTools.", default=False,
+                          dest="preserve_entire_environment")
 
     exgroup = parser.add_mutually_exclusive_group()
     exgroup.add_argument("--rm-container", action="store_true", default=True,
                          help="Delete Docker container used by jobs after they exit (default)",
                          dest="rm_container")
 
-    exgroup.add_argument("--leave-container", action="store_false",
-                         default=True, help="Do not delete Docker container used by jobs after they exit",
-                         dest="rm_container")
+    exgroup.add_argument(
+        "--leave-container", action="store_false", default=True,
+        help="Do not delete Docker container used by jobs after they exit",
+        dest="rm_container")
 
-    cidgroup = parser.add_argument_group("Options for recording the Docker "
-        "container identifier into a file")
+    cidgroup = parser.add_argument_group(
+        "Options for recording the Docker container identifier into a file")
     cidgroup.add_argument("--record-container-id", action="store_true",
                           default=False,
                           help="If enabled, store the Docker container ID into a file. "
                           "See --cidfile-dir to specify the directory.",
                           dest="record_container_id")
 
-    cidgroup.add_argument("--cidfile-dir", type=Text,
-                        help="Directory for storing the Docker container ID file. "
-                             "The default is the current directory",
-                        default="",
-                        dest="cidfile_dir")
+    cidgroup.add_argument(
+        "--cidfile-dir", type=Text, help="Directory for storing the Docker "
+        "container ID file. The default is the current directory",
+        default="", dest="cidfile_dir")
 
-    cidgroup.add_argument("--cidfile-prefix", type=Text,
-                        help="Specify a prefix to the container ID filename. "
-                             "Final file name will be followed by a timestamp. "
-                             "The default is no prefix.",
-                        default="",
-                        dest="cidfile_prefix")
+    cidgroup.add_argument(
+        "--cidfile-prefix", type=Text,
+        help="Specify a prefix to the container ID filename. "
+        "Final file name will be followed by a timestamp. The default is no prefix.",
+        default="", dest="cidfile_prefix")
 
     parser.add_argument("--tmpdir-prefix", type=Text,
                         help="Path prefix for temporary directories",
@@ -79,8 +75,9 @@ def arg_parser():  # type: () -> argparse.ArgumentParser
                          help="Path prefix for intermediate output directories",
                          default=DEFAULT_TMP_PREFIX)
 
-    exgroup.add_argument("--cachedir", type=Text, default="",
-                         help="Directory to cache intermediate workflow outputs to avoid recomputing steps.")
+    exgroup.add_argument(
+        "--cachedir", type=Text, default="",
+        help="Directory to cache intermediate workflow outputs to avoid recomputing steps.")
 
     exgroup = parser.add_mutually_exclusive_group()
     exgroup.add_argument("--rm-tmpdir", action="store_true", default=True,
@@ -92,9 +89,10 @@ def arg_parser():  # type: () -> argparse.ArgumentParser
                          dest="rm_tmpdir")
 
     exgroup = parser.add_mutually_exclusive_group()
-    exgroup.add_argument("--move-outputs", action="store_const", const="move", default="move",
-                         help="Move output files to the workflow output directory and delete intermediate output directories (default).",
-                         dest="move_outputs")
+    exgroup.add_argument(
+        "--move-outputs", action="store_const", const="move", default="move",
+        help="Move output files to the workflow output directory and delete "
+        "intermediate output directories (default).", dest="move_outputs")
 
     exgroup.add_argument("--leave-outputs", action="store_const", const="leave", default="move",
                          help="Leave output files in intermediate output directories.",
@@ -120,6 +118,43 @@ def arg_parser():  # type: () -> argparse.ArgumentParser
                         type=float,
                         default=20)
 
+    provgroup = parser.add_argument_group("Options for recording provenance "
+                                          "information of the execution")
+    provgroup.add_argument("--provenance",
+                           help="Save provenance to specified folder as a "
+                           "Research Object that captures and aggregates "
+                           "workflow execution and data products.",
+                           type=Text)
+
+    provgroup.add_argument("--enable-user-provenance", default=False,
+                           action="store_true",
+                           help="Record user account info as part of provenance.",
+                           dest="user_provenance")
+    provgroup.add_argument("--disable-user-provenance", default=False,
+                           action="store_false",
+                           help="Do not record user account info in provenance.",
+                           dest="user_provenance")
+    provgroup.add_argument("--enable-host-provenance", default=False,
+                           action="store_true",
+                           help="Record host info as part of provenance.",
+                           dest="host_provenance")
+    provgroup.add_argument("--disable-host-provenance", default=False,
+                           action="store_false",
+                           help="Do not record host info in provenance.",
+                           dest="host_provenance")
+    provgroup.add_argument(
+        "--orcid", help="Record user ORCID identifier as part of "
+        "provenance, e.g. https://orcid.org/0000-0002-1825-0097 "
+        "or 0000-0002-1825-0097. Alternatively the environment variable "
+        "ORCID may be set.", dest="orcid", default=os.environ.get("ORCID", ''),
+        type=Text)
+    provgroup.add_argument(
+        "--full-name", help="Record full name of user as part of provenance, "
+        "e.g. Josiah Carberry. You may need to use shell quotes to preserve "
+        "spaces. Alternatively the environment variable CWL_FULL_NAME may "
+        "be set.", dest="cwl_full_name", default=os.environ.get("CWL_FULL_NAME", ''),
+        type=Text)
+
     exgroup = parser.add_mutually_exclusive_group()
     exgroup.add_argument("--print-rdf", action="store_true",
                          help="Print corresponding RDF graph for workflow and exit")
@@ -132,6 +167,10 @@ def arg_parser():  # type: () -> argparse.ArgumentParser
     exgroup.add_argument("--version", action="store_true", help="Print version and exit")
     exgroup.add_argument("--validate", action="store_true", help="Validate CWL document only.")
     exgroup.add_argument("--print-supported-versions", action="store_true", help="Print supported CWL specs.")
+    exgroup.add_argument("--print-subgraph", action="store_true",
+                         help="Print workflow subgraph that will execute "
+                         "(can combine with --target)")
+    exgroup.add_argument("--print-targets", action="store_true", help="Print targets (output parameters)")
 
     exgroup = parser.add_mutually_exclusive_group()
     exgroup.add_argument("--strict", action="store_true",
@@ -148,10 +187,21 @@ def arg_parser():  # type: () -> argparse.ArgumentParser
     exgroup.add_argument("--quiet", action="store_true", help="Only print warnings and errors.")
     exgroup.add_argument("--debug", action="store_true", help="Print even more logging")
 
+    parser.add_argument(
+        "--strict-memory-limit", action="store_true", help="When running with "
+        "software containers and the Docker engine, pass either the "
+        "calculated memory allocation from ResourceRequirements or the "
+        "default of 1 gigabyte to Docker's --memory option.")
+
     parser.add_argument("--timestamps", action="store_true", help="Add "
                         "timestamps to the errors, warnings, and "
                         "notifications.")
     parser.add_argument("--js-console", action="store_true", help="Enable javascript console output")
+    parser.add_argument("--disable-js-validation", action="store_true", help="Disable javascript validation.")
+    parser.add_argument("--js-hint-options-file",
+                        type=Text,
+                        help="File of options to pass to jshint."
+                        "This includes the added option \"includewarnings\". ")
     dockergroup = parser.add_mutually_exclusive_group()
     dockergroup.add_argument("--user-space-docker-cmd", metavar="CMD",
                         help="(Linux/OS X only) Specify a user space docker "
@@ -203,12 +253,12 @@ def arg_parser():  # type: () -> argparse.ArgumentParser
                         help="Specify a default docker container that will be used if the workflow fails to specify one.")
     parser.add_argument("--no-match-user", action="store_true",
                         help="Disable passing the current uid to `docker run --user`")
-    parser.add_argument("--disable-net", action="store_true",
-                        help="Use docker's default networking for containers;"
-                             " the default is to enable networking.")
     parser.add_argument("--custom-net", type=Text,
-                        help="Will be passed to `docker run` as the '--net' "
-                             "parameter. Implies '--enable-net'.")
+                        help="Passed to `docker run` as the '--net' "
+                             "parameter when NetworkAccess is true.")
+    parser.add_argument("--disable-validate", dest="do_validate",
+                        action="store_false", default=True,
+                        help=argparse.SUPPRESS)
 
     exgroup = parser.add_mutually_exclusive_group()
     exgroup.add_argument("--enable-ga4gh-tool-registry", action="store_true", help="Enable resolution using GA4GH tool registry API",
@@ -247,6 +297,10 @@ def arg_parser():  # type: () -> argparse.ArgumentParser
     parser.add_argument("--overrides", type=str,
                         default=None, help="Read process requirement overrides from file.")
 
+    parser.add_argument("--target", "-t", action="append",
+                        help="Only execute steps that contribute to "
+                        "listed targets (can provide more than once).")
+
     parser.add_argument("workflow", type=Text, nargs="?", default=None,
             metavar='cwl_document', help="path or URL to a CWL Workflow, "
             "CommandLineTool, or ExpressionTool. If the `inputs_object` has a "
@@ -266,7 +320,7 @@ def get_default_args():
     Get default values of cwltool's command line options
     """
     ap = arg_parser()
-    args = ap.parse_args()
+    args = ap.parse_args([])
     return vars(args)
 
 
@@ -335,8 +389,8 @@ def add_argument(toolparser, name, inptype, records, description="",
     else:
         flag = "--"
 
-    required = True
-    if isinstance(inptype, list):
+    required = default is None
+    if isinstance(inptype, MutableSequence):
         if inptype[0] == "null":
             required = False
             if len(inptype) == 2:
@@ -346,23 +400,23 @@ def add_argument(toolparser, name, inptype, records, description="",
                 return None
 
     ahelp = description.replace("%", "%%")
-    action = None  # type: Union[argparse.Action, Text]
+    action = None  # type: Optional[Union[argparse.Action, Text]]
     atype = None  # type: Any
 
     if inptype == "File":
         action = cast(argparse.Action, FileAction)
     elif inptype == "Directory":
         action = cast(argparse.Action, DirectoryAction)
-    elif isinstance(inptype, dict) and inptype["type"] == "array":
+    elif isinstance(inptype, MutableMapping) and inptype["type"] == "array":
         if inptype["items"] == "File":
             action = cast(argparse.Action, FileAppendAction)
         elif inptype["items"] == "Directory":
             action = cast(argparse.Action, DirectoryAppendAction)
         else:
             action = "append"
-    elif isinstance(inptype, dict) and inptype["type"] == "enum":
+    elif isinstance(inptype, MutableMapping) and inptype["type"] == "enum":
         atype = Text
-    elif isinstance(inptype, dict) and inptype["type"] == "record":
+    elif isinstance(inptype, MutableMapping) and inptype["type"] == "record":
         records.append(name)
         for field in inptype['fields']:
             fieldname = name + "." + shortname(field['name'])
@@ -372,7 +426,7 @@ def add_argument(toolparser, name, inptype, records, description="",
                 toolparser, fieldname, fieldtype, records,
                 fielddescription)
         return
-    if inptype == "string":
+    elif inptype == "string":
         atype = Text
     elif inptype == "int":
         atype = int
@@ -382,11 +436,7 @@ def add_argument(toolparser, name, inptype, records, description="",
         atype = float
     elif inptype == "boolean":
         action = "store_true"
-
-    if default:
-        required = False
-
-    if not atype and not action:
+    else:
         _logger.debug(u"Can't make command line argument from %s", inptype)
         return None
 
