@@ -1,34 +1,40 @@
 """Loads a CWL document."""
 from __future__ import absolute_import
-# pylint: disable=unused-import
 
+import hashlib
 import logging
 import os
 import re
 import uuid
-import hashlib
-import json
-import copy
-from typing import (Any, Callable, Dict, Iterable, List, Mapping, Optional,
-        Text, Tuple, Union, cast)
+from typing import (Any, Callable, Dict, List, MutableMapping, MutableSequence,
+                    Optional, Tuple, Union, cast)
 
 import requests.sessions
+from ruamel.yaml.comments import CommentedMap, CommentedSeq
+from schema_salad import schema
+from schema_salad.ref_resolver import (ContextType,  # pylint: disable=unused-import
+                                       Fetcher, Loader, file_uri)
+from schema_salad.sourceline import SourceLine, cmap
+from schema_salad.validate import ValidationException
 from six import itervalues, string_types
 from six.moves import urllib
-
-import schema_salad.schema as schema
-from avro.schema import Names
-from ruamel.yaml.comments import CommentedMap, CommentedSeq
-from schema_salad.ref_resolver import ContextType, Fetcher, Loader, file_uri
-from schema_salad.sourceline import cmap, SourceLine
-from schema_salad.validate import ValidationException
+from typing_extensions import Text  # pylint: disable=unused-import
+# move to a regular typing import when Python 3.3-3.6 is no longer supported
 
 from . import process, update
+from .context import LoadingContext  # pylint: disable=unused-import
 from .errors import WorkflowException
-from .process import Process, shortname, get_schema
+from .loghandler import _logger
+from .process import (Process, get_schema,  # pylint: disable=unused-import
+                      shortname)
+from .software_requirements import (  # pylint: disable=unused-import
+    DependenciesConfiguration)
 from .update import ALLUPDATES
+from .utils import json_dumps
+
+
+
 
-_logger = logging.getLogger("cwltool")
 jobloaderctx = {
     u"cwl": "https://w3id.org/cwl/cwl#",
     u"cwltool": "http://commonwl.org/cwltool#",
@@ -53,34 +59,34 @@ overrides_ctx = {
 }  # type: ContextType
 
 
-FetcherConstructorType = Callable[[Dict[Text, Union[Text, bool]],
-    requests.sessions.Session], Fetcher]
+FetcherConstructorType = Callable[
+    [Dict[Text, Union[Text, bool]], requests.sessions.Session], Fetcher]
+ResolverType = Callable[[Loader, Union[Text, Dict[Text, Any]]], Text]
 
-loaders = {}  # type: Dict[FetcherConstructorType, Loader]
+loaders = {}  # type: Dict[Optional[FetcherConstructorType], Loader]
 
 def default_loader(fetcher_constructor):
     # type: (Optional[FetcherConstructorType]) -> Loader
     if fetcher_constructor in loaders:
         return loaders[fetcher_constructor]
-    else:
-        loader = Loader(jobloaderctx, fetcher_constructor=fetcher_constructor)
-        loaders[fetcher_constructor] = loader
-        return loader
+    loader = Loader(jobloaderctx, fetcher_constructor=fetcher_constructor)
+    loaders[fetcher_constructor] = loader
+    return loader
 
 def resolve_tool_uri(argsworkflow,  # type: Text
-                     resolver=None,  # type: Callable[[Loader, Union[Text, Dict[Text, Any]]], Text]
+                     resolver=None,  # type: ResolverType
                      fetcher_constructor=None,  # type: FetcherConstructorType
                      document_loader=None  # type: Loader
-                     ):  # type: (...) -> Tuple[Text, Text]
+                    ):  # type: (...) -> Tuple[Text, Text]
 
-    uri = None  # type: Text
+    uri = None  # type: Optional[Text]
     split = urllib.parse.urlsplit(argsworkflow)
     # In case of Windows path, urlsplit misjudge Drive letters as scheme, here we are skipping that
     if split.scheme and split.scheme in [u'http', u'https', u'file']:
         uri = argsworkflow
     elif os.path.exists(os.path.abspath(argsworkflow)):
         uri = file_uri(str(os.path.abspath(argsworkflow)))
-    elif resolver:
+    elif resolver is not None:
         if document_loader is None:
             document_loader = default_loader(fetcher_constructor)  # type: ignore
         uri = resolver(document_loader, argsworkflow)
@@ -103,17 +109,18 @@ def fetch_document(argsworkflow,  # type: Union[Text, Dict[Text, Any]]
 
     document_loader = default_loader(fetcher_constructor)  # type: ignore
 
-    uri = None  # type: Text
-    workflowobj = None  # type: CommentedMap
+    uri = None  # type: Optional[Text]
+    workflowobj = None  # type: Optional[CommentedMap]
     if isinstance(argsworkflow, string_types):
         uri, fileuri = resolve_tool_uri(argsworkflow, resolver=resolver,
-                document_loader=document_loader)
+                                        document_loader=document_loader)
         workflowobj = document_loader.fetch(fileuri)
-    elif isinstance(argsworkflow, dict):
+    elif isinstance(argsworkflow, MutableMapping):
         uri = "#" + Text(id(argsworkflow))
         workflowobj = cast(CommentedMap, cmap(argsworkflow, fn=uri))
     else:
         raise ValidationException("Must be URI or object: '%s'" % argsworkflow)
+    assert workflowobj is not None
 
     return document_loader, workflowobj, uri
 
@@ -121,26 +128,36 @@ def fetch_document(argsworkflow,  # type: Union[Text, Dict[Text, Any]]
 def _convert_stdstreams_to_files(workflowobj):
     # type: (Union[Dict[Text, Any], List[Dict[Text, Any]]]) -> None
 
-    if isinstance(workflowobj, dict):
+    if isinstance(workflowobj, MutableMapping):
         if workflowobj.get('class') == 'CommandLineTool':
-            for out in workflowobj.get('outputs', []):
-                if type(out) is not CommentedMap:
-                    with SourceLine(workflowobj, "outputs", ValidationException, _logger.isEnabledFor(logging.DEBUG)):
-                        raise ValidationException("Output '%s' is not a valid OutputParameter." % out)
-                for streamtype in ['stdout', 'stderr']:
-                    if out.get('type') == streamtype:
-                        if 'outputBinding' in out:
-                            raise ValidationException(
-                                "Not allowed to specify outputBinding when"
-                                " using %s shortcut." % streamtype)
-                        if streamtype in workflowobj:
-                            filename = workflowobj[streamtype]
-                        else:
-                            filename = Text(hashlib.sha1(json.dumps(workflowobj,
-                                        sort_keys=True).encode('utf-8')).hexdigest())
-                            workflowobj[streamtype] = filename
-                        out['type'] = 'File'
-                        out['outputBinding'] = cmap({'glob': filename})
+            with SourceLine(workflowobj, "outputs", ValidationException,
+                            _logger.isEnabledFor(logging.DEBUG)):
+                outputs = workflowobj.get('outputs', [])
+                if not isinstance(outputs, CommentedSeq):
+                    raise ValidationException('"outputs" section is not '
+                                              'valid.')
+                for out in workflowobj.get('outputs', []):
+                    if not isinstance(out, CommentedMap):
+                        raise ValidationException(
+                            "Output '{}' is not a valid "
+                            "OutputParameter.".format(out))
+                    for streamtype in ['stdout', 'stderr']:
+                        if out.get('type') == streamtype:
+                            if 'outputBinding' in out:
+                                raise ValidationException(
+                                    "Not allowed to specify outputBinding when"
+                                    " using %s shortcut." % streamtype)
+                            if streamtype in workflowobj:
+                                filename = workflowobj[streamtype]
+                            else:
+                                filename = Text(
+                                    hashlib.sha1(json_dumps(workflowobj,
+                                                            sort_keys=True
+                                                           ).encode('utf-8')
+                                                ).hexdigest())
+                                workflowobj[streamtype] = filename
+                            out['type'] = 'File'
+                            out['outputBinding'] = cmap({'glob': filename})
             for inp in workflowobj.get('inputs', []):
                 if inp.get('type') == 'stdin':
                     if 'inputBinding' in inp:
@@ -159,51 +176,53 @@ def _convert_stdstreams_to_files(workflowobj):
         else:
             for entry in itervalues(workflowobj):
                 _convert_stdstreams_to_files(entry)
-    if isinstance(workflowobj, list):
+    if isinstance(workflowobj, MutableSequence):
         for entry in workflowobj:
             _convert_stdstreams_to_files(entry)
 
 def _add_blank_ids(workflowobj):
     # type: (Union[Dict[Text, Any], List[Dict[Text, Any]]]) -> None
 
-    if isinstance(workflowobj, dict):
+    if isinstance(workflowobj, MutableMapping):
         if ("run" in workflowobj and
-            isinstance(workflowobj["run"], dict) and
-            "id" not in workflowobj["run"] and
-            "$import" not in workflowobj["run"]):
+                isinstance(workflowobj["run"], MutableMapping) and
+                "id" not in workflowobj["run"] and
+                "$import" not in workflowobj["run"]):
             workflowobj["run"]["id"] = Text(uuid.uuid4())
         for entry in itervalues(workflowobj):
             _add_blank_ids(entry)
-    if isinstance(workflowobj, list):
+    if isinstance(workflowobj, MutableSequence):
         for entry in workflowobj:
             _add_blank_ids(entry)
 
-def validate_document(document_loader,  # type: Loader
-                      workflowobj,  # type: CommentedMap
-                      uri,  # type: Text
-                      enable_dev=False,  # type: bool
-                      strict=True,  # type: bool
-                      preprocess_only=False,  # type: bool
+def validate_document(document_loader,           # type: Loader
+                      workflowobj,               # type: CommentedMap
+                      uri,                       # type: Text
+                      overrides,                 # type: List[Dict]
+                      metadata,                  # type: Dict[Text, Any]
+                      enable_dev=False,          # type: bool
+                      strict=True,               # type: bool
+                      preprocess_only=False,     # type: bool
                       fetcher_constructor=None,  # type: FetcherConstructorType
-                      skip_schemas=None,  # type: bool
-                      overrides=None,  # type: List[Dict]
-                      metadata=None,  # type: Optional[Dict]
-                      ):
-    # type: (...) -> Tuple[Loader, Names, Union[Dict[Text, Any], List[Dict[Text, Any]]], Dict[Text, Any], Text]
+                      skip_schemas=None,         # type: bool
+                      do_validate=True           # type: bool
+                     ):
+    # type: (...) -> Tuple[Loader, schema.Names, Union[Dict[Text, Any], List[Dict[Text, Any]]], Dict[Text, Any], Text]
     """Validate a CWL document."""
 
-    if isinstance(workflowobj, list):
+    if isinstance(workflowobj, MutableSequence):
         workflowobj = cmap({
             "$graph": workflowobj
         }, fn=uri)
 
-    if not isinstance(workflowobj, dict):
-        raise ValueError("workflowjobj must be a dict, got '%s': %s" % (type(workflowobj), workflowobj))
+    if not isinstance(workflowobj, MutableMapping):
+        raise ValueError("workflowjobj must be a dict, got '{}': {}".format(
+            type(workflowobj), workflowobj))
 
     jobobj = None
     if "cwl:tool" in workflowobj:
         job_loader = default_loader(fetcher_constructor)  # type: ignore
-        jobobj, _ = job_loader.resolve_all(workflowobj, uri)
+        jobobj, _ = job_loader.resolve_all(workflowobj, uri, checklinks=do_validate)
         uri = urllib.parse.urljoin(uri, workflowobj["https://w3id.org/cwl/cwl#tool"])
         del cast(dict, jobobj)["https://w3id.org/cwl/cwl#tool"]
 
@@ -215,16 +234,21 @@ def validate_document(document_loader,  # type: Loader
 
     fileuri = urllib.parse.urldefrag(uri)[0]
     if "cwlVersion" not in workflowobj:
-        if metadata and 'cwlVersion' in metadata:
+        if 'cwlVersion' in metadata:
             workflowobj['cwlVersion'] = metadata['cwlVersion']
         else:
             raise ValidationException(
-                  "No cwlVersion found. "
-                  "Use the following syntax in your CWL document to declare the version: cwlVersion: <version>.\n"
-                  "Note: if this is a CWL draft-2 (pre v1.0) document then it will need to be upgraded first.")
+                "No cwlVersion found. "
+                "Use the following syntax in your CWL document to declare "
+                "the version: cwlVersion: <version>.\n"
+                "Note: if this is a CWL draft-2 (pre v1.0) document then it "
+                "will need to be upgraded first.")
 
-    if not isinstance(workflowobj["cwlVersion"], (str, Text)):
-        raise Exception("'cwlVersion' must be a string, got %s" % type(workflowobj["cwlVersion"]))
+    if not isinstance(workflowobj["cwlVersion"], string_types):
+        with SourceLine(workflowobj, "cwlVersion", ValidationException):
+            raise ValidationException("'cwlVersion' must be a string, "
+                                      "got {}".format(
+                                          type(workflowobj["cwlVersion"])))
     # strip out version
     workflowobj["cwlVersion"] = re.sub(
         r"^(?:cwl:|https://w3id.org/cwl/cwl#)", "",
@@ -237,8 +261,10 @@ def validate_document(document_loader,  # type: Loader
                 version += " (with --enable-dev flag only)"
             versions.append(version)
         versions.sort()
-        raise ValidationException("The CWL reference runner no longer supports pre CWL v1.0 documents. "
-                                  "Supported versions are: \n{}".format("\n".join(versions)))
+        raise ValidationException(
+            "The CWL reference runner no longer supports pre CWL v1.0 "
+            "documents. Supported versions are: "
+            "\n{}".format("\n".join(versions)))
 
     (sch_document_loader, avsc_names) = \
         process.get_schema(workflowobj["cwlVersion"])[:2]
@@ -254,7 +280,8 @@ def validate_document(document_loader,  # type: Loader
     _add_blank_ids(workflowobj)
 
     workflowobj["id"] = fileuri
-    processobj, new_metadata = document_loader.resolve_all(workflowobj, fileuri)
+    processobj, new_metadata = document_loader.resolve_all(
+        workflowobj, fileuri, checklinks=do_validate)
     if not isinstance(processobj, (CommentedMap, CommentedSeq)):
         raise ValidationException("Workflow must be a dict or list.")
 
@@ -269,13 +296,14 @@ def validate_document(document_loader,  # type: Loader
     if preprocess_only:
         return document_loader, avsc_names, processobj, new_metadata, uri
 
-    schema.validate_doc(avsc_names, processobj, document_loader, strict)
+    if do_validate:
+        schema.validate_doc(avsc_names, processobj, document_loader, strict)
 
     if new_metadata.get("cwlVersion") != update.LATEST:
         processobj = cast(CommentedMap, cmap(update.update(
             processobj, document_loader, fileuri, enable_dev, new_metadata)))
 
-    if jobobj:
+    if jobobj is not None:
         new_metadata[u"cwl:defaults"] = jobobj
 
     if overrides:
@@ -284,19 +312,17 @@ def validate_document(document_loader,  # type: Loader
     return document_loader, avsc_names, processobj, new_metadata, uri
 
 
-def make_tool(document_loader,  # type: Loader
-              avsc_names,  # type: Names
-              metadata,  # type: Dict[Text, Any]
-              uri,  # type: Text
-              makeTool,  # type: Callable[..., Process]
-              kwargs  # type: dict
-              ):
-    # type: (...) -> Process
+def make_tool(document_loader,    # type: Loader
+              avsc_names,         # type: schema.Names
+              metadata,           # type: Dict[Text, Any]
+              uri,                # type: Union[Text, CommentedMap, CommentedSeq]
+              loadingContext      # type: LoadingContext
+             ):  # type: (...) -> Process
     """Make a Python CWL object."""
     resolveduri = document_loader.resolve_ref(uri)[0]
 
     processobj = None
-    if isinstance(resolveduri, list):
+    if isinstance(resolveduri, MutableSequence):
         for obj in resolveduri:
             if obj['id'].endswith('#main'):
                 processobj = obj
@@ -307,19 +333,17 @@ def make_tool(document_loader,  # type: Loader
                 "one of #%s" % ", #".join(
                     urllib.parse.urldefrag(i["id"])[1] for i in resolveduri
                     if "id" in i))
-    elif isinstance(resolveduri, dict):
+    elif isinstance(resolveduri, MutableMapping):
         processobj = resolveduri
     else:
         raise Exception("Must resolve to list or dict")
 
-    kwargs = kwargs.copy()
-    kwargs.update({
-        "makeTool": makeTool,
-        "loader": document_loader,
-        "avsc_names": avsc_names,
-        "metadata": metadata
-    })
-    tool = makeTool(processobj, **kwargs)
+    loadingContext = loadingContext.copy()
+    loadingContext.loader = document_loader
+    loadingContext.avsc_names = avsc_names
+    loadingContext.metadata = metadata
+
+    tool = loadingContext.construct_tool_object(processobj, loadingContext)
 
     if "cwl:defaults" in metadata:
         jobobj = metadata["cwl:defaults"]
@@ -330,28 +354,33 @@ def make_tool(document_loader,  # type: Loader
     return tool
 
 
-def load_tool(argsworkflow,  # type: Union[Text, Dict[Text, Any]]
-              makeTool,  # type: Callable[..., Process]
-              kwargs=None,  # type: Dict
-              enable_dev=False,  # type: bool
-              strict=True,  # type: bool
-              resolver=None,  # type: Callable[[Loader, Union[Text, Dict[Text, Any]]], Text]
-              fetcher_constructor=None,  # type: FetcherConstructorType
-              overrides=None
-              ):
-    # type: (...) -> Process
+def load_tool(argsworkflow,              # type: Union[Text, Dict[Text, Any]]
+              loadingContext             # type: LoadingContext
+             ):  # type: (...) -> Process
 
-    document_loader, workflowobj, uri = fetch_document(argsworkflow, resolver=resolver,
-                                                       fetcher_constructor=fetcher_constructor)
-    document_loader, avsc_names, processobj, metadata, uri = validate_document(
-        document_loader, workflowobj, uri, enable_dev=enable_dev,
-        strict=strict, fetcher_constructor=fetcher_constructor,
-        overrides=overrides, metadata=kwargs.get('metadata', None)
-        if kwargs else None)
-    return make_tool(document_loader, avsc_names, metadata, uri,
-                     makeTool, kwargs if kwargs else {})
+    document_loader, workflowobj, uri = fetch_document(
+        argsworkflow,
+        resolver=loadingContext.resolver,
+        fetcher_constructor=loadingContext.fetcher_constructor)
 
-def resolve_overrides(ov, ov_uri, baseurl):  # type: (CommentedMap, Text, Text) -> List[Dict[Text, Any]]
+    document_loader, avsc_names, _, metadata, uri = validate_document(
+        document_loader, workflowobj, uri,
+        enable_dev=loadingContext.enable_dev,
+        strict=loadingContext.strict,
+        fetcher_constructor=loadingContext.fetcher_constructor,
+        overrides=loadingContext.overrides_list,
+        metadata=loadingContext.metadata)
+
+    return make_tool(document_loader,
+                     avsc_names,
+                     metadata,
+                     uri,
+                     loadingContext)
+
+def resolve_overrides(ov,      # Type: CommentedMap
+                      ov_uri,  # Type: Text
+                      baseurl  # type: Text
+                     ):  # type: (...) -> List[Dict[Text, Any]]
     ovloader = Loader(overrides_ctx)
     ret, _ = ovloader.resolve_all(ov, baseurl)
     if not isinstance(ret, CommentedMap):
